from maldives.api import FredData

import yfinance as yf
import pandas as pd

# TODO: use gold price as feature and silver price as target
def load_silver_data(start=pd.Timestamp.today()-pd.tseries.offsets.BDay(300), end=pd.Timestamp.today()):
    if type(start) is int:
            start = end - pd.tseries.offsets.BDay(start)
    pass     

def load_gold_data(fred_api_key, start=pd.Timestamp.today()-pd.tseries.offsets.BDay(300), end=pd.Timestamp.today()):
    if type(start) is int:
            start = end - pd.tseries.offsets.BDay(start)

    # load gold futures prices
    gold = yf.Ticker("GC=F").history(start=start, end=end)
    gold.index = gold.index.date
    gold['ClosingPrice'] = gold['Close']
    gold = gold[['ClosingPrice']]

    # load cpi and treasury yield
    fred = FredData(fred_api_key)
    cpi, treasury = fred.CPI(), fred.Treasury10Y()

    df = pd.merge_asof(gold.join(
        treasury), cpi, left_index=True, right_index=True, direction='nearest')
    df = df.dropna().loc[start.date():end.date()]
    return df

def load_silver_data(fred_api_key, start=pd.Timestamp.today()-pd.tseries.offsets.BDay(300), end=pd.Timestamp.today()):
    if type(start) is int:
            start = end - pd.tseries.offsets.BDay(start)

    # load gold futures prices
    gold = yf.Ticker("GC=F").history(start=start, end=end)
    gold.index = gold.index.date
    gold['ClosingPrice'] = gold['Close']
    gold = gold[['ClosingPrice']]

    # load silver futures prices
    silver = yf.Ticker("SI=F").history(start=start, end=end)
    silver.index = silver.index.date
    silver['ClosingPrice'] = silver['Close']
    silver = silver[['ClosingPrice']]

    df = gold.join(silver)
    df = df.dropna().loc[start.date():end.date()]
    return df


# %%
import numpy as np
import pandas as pd
import re
from zipfile import ZipFile
from maldives.utils.scrapping import download_links_on_page


class USDACattleData(object):
    def __init__(self, base_url="https://usda.library.cornell.edu/concern/publications/m326m174z", work_dir="cattle_data"):
        self.base_url = base_url
        self.work_dir = work_dir
        self.df = None

    # Methods for parsing
    @staticmethod
    def _extract_tables(lines, titles):
        tables = {k: [] for k in titles.keys()}
        write_to = None  # pointer to which table we are writing

        for ln in lines:
            if type(ln) is bytes:
                ln = ln.decode('latin1')
            ln = ln.strip()

            if write_to is not None:
                if ('"d"' in ln) or (',d,' in ln):
                    write_to.append(ln.strip().strip(','))
                elif ('"c"' in ln) or (',c,' in ln):
                    write_to = None
            else:
                for k, v in titles.items():
                    if any([vv in ln for vv in v]) and len(tables[k]) == 0:
                        write_to = tables[k]
                        break
        return tables

    @staticmethod
    def _parse_zipped_data(fname):
        zf = ZipFile(fname)
        table_titles = {'Placed': ['Cattle Placed on Feed by Weight Group'],
                        'Inventory': ['Cattle on Feed Inventory, Placements, Marketings, and Other Disappearance',
                                      'Cattle on Feed, Placements, Marketings, and Other Disappearance'],
                        'ByClass': ['on Feed by Class']}
        tables = USDACattleData._extract_tables(
            zf.open('cofd_all_tables.csv').readlines(), table_titles)

        FORMAT_ONFEEDBYWEIGHT = {'PLACED_U600': 3, 'PLACED_U700': 4, 'PLACED_U800': 5,
                                 'PLACED_U900': 6, 'PLACED_U1000': 8, 'PLACED_A1000': 9}

        def _format_tuple(tuple, format):
            return {k: tuple[v] for k, v in format.items()}

        data = _format_tuple(
            tables['Placed'][-1].split(','), FORMAT_ONFEEDBYWEIGHT)
        data['ONFEED_OPEN'] = tables['Inventory'][0].split(',')[-2]
        data['PLACED'] = tables['Inventory'][1].split(',')[-2]
        data['MARKETED'] = tables['Inventory'][2].split(',')[-2]
        data['DISAPPEAR'] = tables['Inventory'][3].split(',')[-2]
        data['ONFEED_CLOSE'] = tables['Inventory'][4].split(',')[-2]

        data['ONFEED_STEER'] = tables['ByClass'][1].split(
            ',')[-3] if len(tables['ByClass']) > 0 else np.nan
        data['ONFEED_HEIFER'] = tables['ByClass'][2].split(
            ',')[-3] if len(tables['ByClass']) > 0 else np.nan

        return data

    @staticmethod
    def _parse_date(fname):
        fname = fname.split('/')[-1]
        if fname.startswith('cofd'):  # new format
            date_format = ['\d{4}', '%m%y']
        else:
            date_format = ['\d{2}-\d{2}-\d{4}', '%m-%d-%Y']
        return pd.to_datetime(re.search(date_format[0], fname).group(), format=date_format[1]).replace(day=1)

    def load_data(self, num_months=60, overwrite=False):
        data_files = []
        page = 1
        while len(data_files) < num_months:
            data_files.extend(download_links_on_page(
                f"{self.base_url}?page={page}", self.work_dir, overwrite=overwrite, regex="*.zip"))
            page += 1

        entries = []
        for fname in set(data_files):
            try:
                data = USDACattleData._parse_zipped_data(fname)
                data['Date'] = USDACattleData._parse_date(
                    fname)-pd.DateOffset(months=1)  # publication date is one-month behind data date
                entries.append(data)
            except Exception as ex:
                print(f"Unable to parse {fname}.")
                print(ex)
                continue
        self.df = pd.DataFrame(entries).set_index(
            'Date').sort_index().astype(float)

        import_export = BeefTradeData().load_data()
        self.df = self.df.join(import_export[['NETIMPORT']])
        return self.df

    def info(self):
        columns_descriptions = {
            'PLACED_U600': 'Cattles under 600 lbs placed on feed during the month [Unit: 1000 head]',
            'PLACED_U700': 'Cattles of 600-699 lbs placed on feed during the month [Unit: 1000 head]',
            'PLACED_U800': 'Cattles of 700-799 lbs placed on feed during the month [Unit: 1000 head]',
            'PLACED_U900': 'Cattles of 800-899 lbs placed on feed during the month [Unit: 1000 head]',
            'PLACED_U1000': 'Cattles of 900-1000 lbs placed on feed during the month [Unit: 1000 head]',
            'PLACED_A1000': 'Cattles above 1000 lbs placed on feed during the month [Unit: 1000 head]',
            'PLACED': 'Total number of cattles placed on feed during the month [Unit: 1000 head]',
            'ONFEED_OPEN': 'Cattles that are on feed at the beginning of the month [Unit: 1000 head]',
            'ONFEED_CLOSE': 'Cattles that are on feed at the end of the month [Unit: 1000 head]',
            'MARKETED': 'Cattles that are slaughtered during the month [Unit: 1000 head]',
            'DISAPPEAR': 'Cattles that disappeared during the month [Unit: 1000 head]',
            'NETIMPORT': 'Net import of beef [Unit:KG]'
        }
        return pd.Series(columns_descriptions)


class BeefTradeData(object):
    def __init__(self, base_url="https://www.ers.usda.gov/data-products/livestock-and-meat-international-trade-data", work_dir="meat_trade_data"):
        self.base_url = base_url
        self.work_dir = work_dir
        self.df = None

    @staticmethod
    def _format_df(df, desc='bovine animals'):
        df['Date'] = pd.to_datetime(df.YEAR_ID.astype(
            str)+'/'+df.TIMEPERIOD_ID.astype(str)+'/'+'01')

        bovine = df.COMMODITY_DESC.str.contains(desc)
        return df[bovine].query('UNIT_DESC=="KG" and GEOGRAPHY_DESC=="World"').groupby(['Date']).AMOUNT.sum()

    def load_data(self, overwrite=True):
        download_links_on_page(self.base_url, self.work_dir,
                               overwrite=overwrite, regex='*.zip?v*')

        zf = ZipFile(f'{self.work_dir}/LivestockMeatTrade.zip')
        df_export = BeefTradeData._format_df(
            pd.read_csv(zf.open('LivestockMeat_Exports.csv')))
        df_import = BeefTradeData._format_df(
            pd.read_csv(zf.open('LivestockMeat_Imports.csv')))

        self.df = pd.concat([df_export.rename('Export'),
                            df_import.rename('Import')], axis=1)
        self.df['NETIMPORT'] = self.df.eval('Import-Export')

        return self.df


class CattleWeeklyProduction(object):
    def __init__(self, base_url="https://usda.library.cornell.edu/concern/publications/2j62s4898", work_dir="production_data"):
        self.base_url = base_url
        self.work_dir = work_dir
        self.df = None

    @staticmethod
    def _extract_tables(lines, titles):
        tables = {k: [] for k in titles.keys()}
        write_to = None  # pointer to which table we are writing

        for ln in lines:
            if type(ln) is bytes:
                ln = ln.decode('latin1')
            ln = ln.strip().strip('-')

            if write_to is not None:
                if len(ln) == 0:
                    write_to = None
                else:
                    write_to.append(ln)
            else:
                for k, v in titles.items():
                    if any([vv in ln for vv in v]) and len(tables[k]) == 0:
                        write_to = tables[k]
                        break
        return tables

    @staticmethod
    def _format_tables(table):
        columns = [c.strip().replace(' ', '_').replace('/', '')
                   for c in table[0].split('  ') if c.strip() != '']
        data = [c.strip().replace(',', '').replace('%', '').replace('*', '')
                for c in table[1].split('  ') if c.strip() != '']
        return {k: v for k, v in zip(columns, data)}

    def load_data(self, num_weeks=60, overwrite=False):
        data_files = []
        page = 1
        while len(data_files) < num_weeks:
            data_files.extend(download_links_on_page(
                f"{self.base_url}?page={page}", self.work_dir, overwrite=overwrite, regex="*.txt", fname_keep_url_levels=3))
            page += 1

        slaughter = []
        byclass = []
        for fname in set(data_files):
            try:
                table_titles = {'Slaughter': ['Livestock Slaughter'],
                                'ByClass': ['Cattle Slaughtered by Class']}
                with open(fname) as f:
                    tables = CattleWeeklyProduction._extract_tables(
                        f.readlines(), table_titles)

                slaughter.append(
                    CattleWeeklyProduction._format_tables(tables['Slaughter']))
                byclass.append(
                    CattleWeeklyProduction._format_tables(tables['ByClass']))
            except Exception as ex:
                print(f"Unable to parse {fname}.")
                print(ex)
                continue

        def clean_df(entries):
            df = pd.DataFrame(entries)
            df['Date'] = pd.to_datetime(df['Week_Ending'], format='%d-%b-%y')
            df = df.set_index('Date').sort_index().drop(
                columns=['Week_Ending'])
            return df.apply(pd.to_numeric, errors='coerce').astype(float)

        return (clean_df(slaughter)/1000).join(clean_df(byclass)/100, how='outer')

    def info(self):
        columns_descriptions = {
            'Cattle': 'Cattle slaughtered [Unit: 1000 head]',
            'CalvesVealer': 'Calves/Vealer slaughtered [Unit: 1000 head]',
            'Hogs': 'Hogs slaughtered [Unit: 1000 head]',
            'SheepLambs': 'Sheeps/Lambs slaughtered [Unit: 1000 head]',
            'Steers': 'Percentage of steers slaughtered',
            'Heifers': 'Percentage of heifers slaughtered',
            'Cows': 'Percentage of cows slaughtered',
            'Bulls': 'Percentage of bulls slaughtered',
        }
        return pd.Series(columns_descriptions)



import pandas as pd
import numpy as np
import plotly.express as px

from sklearn.linear_model import HuberRegressor
from sklearn.metrics import r2_score

from maldives.backtest.utils import CalculateProbability,  VisualizeProbability


def _shift_data(X, y, n):
    # +ve: features predict previous value, -ve: features predict next value
    if n == 0:
        return X, y
    return (X[n:], y[:-n]) if n > 0 else (X[:n], y[-n:])


def _add_lagged_y_to_x(X, y, n):
    # TODO add lagged response to feature, n is (list of) lag
    pass

class LinearRegressor(object):
    def __init__(self, fitter=HuberRegressor(), transform=None, invtransform=None):
        """Basic linear regression model

        Args:
            fitter (scikit-learn model, optional): Model object with fit(), predict() and score(). Defaults to HuberRegressor().
            transform (Function, optional): Transformation applied to data, e.g. log. Defaults to None.
            invtransform (Function, optional): Function to invert transform, e.g. exp. Defaults to None.
        """
        self.fitter = fitter
        identity = (lambda x: x)
        self.transform = identity if transform is None else transform
        self.invtransform = identity if invtransform is None else invtransform
        self.prediction = None
        self.observed = None
        self.R2 = 0
        self.shift = 0

    def _preprocess(self, X, y):
        X_shifted, y_shifted = _shift_data(X, y, self.shift)
        self.observed = y_shifted
        self.X_transformed = self.transform(X_shifted)
        self.y_transformed = self.transform(y_shifted)

    def fit(self, X, y, shift=0, **kwargs):
        """Fit linear model to data

        Args:
            X (array): features
            y (array): response
            shift (int, optional): shift in y. Positive means current features are used to predict previous response. 
            Negative means current features are used to predict next response. Defaults to 0.
        """
        self.shift = shift
        self._preprocess(X, y)
        self.fitter.fit(self.X_transformed, self.y_transformed, **kwargs)
        self.ypred = self.invtransform(self.fitter.predict(self.X_transformed))
        self.R2 = r2_score(self.observed, self.ypred)

    def rolling_fit(self, X, y, shift=0, window=1, **kwargs):
        """Run a rolling fit prediction over the dataset

        Args:
            X (array): features
            y (array): response
            shift (int, optional): shift in y. Positive means current features are used to predict previous response. 
            Negative means current features are used to predict next response. Defaults to 0.
            window (int, optional): Size of the rolling window. Defaults to 1.
        """
        self.shift = shift
        self._preprocess(X, y)

        # loop over windows
        ypred = []
        R2 = []
        coeffs = []
        ind = []
        for start_ind in range(len(self.y_transformed)-window):
            Xw = self.X_transformed[start_ind:start_ind+window]
            yw = self.y_transformed[start_ind:start_ind+window]
            self.fitter.fit(Xw, yw, **kwargs)
            ypredw = self.invtransform(self.fitter.predict(Xw))
            ypred.append(ypredw[-1])
            R2.append(r2_score(self.invtransform(yw), ypredw))
            coeffs.append(self.fitter.coef_)
            if hasattr(Xw, 'index'):
                ind.append(Xw.index[-1])

        results = pd.DataFrame(coeffs)
        results['Prediction'] = ypred
        results['R2'] = R2
        if len(ind) > 0:
            results.index = ind
        return results

    def predict(self, X, **kwargs):
        """Predict on feature.

        Args:
            X (array): features

        Returns:
            prediction (array): predicted response
        """
        self.prediction = self.invtransform(
            self.fitter.predict(self.transform(X), **kwargs))
        return self.prediction


class RegressionModel(object):

    def __init__(self, df, features, target, shift=30):
        """Interface for regression model.

        Args:
            df (pd.Dataframe): pandas dataframe containing the dataset
            features (list of str): column names for features
            target (str): column name for target
            shift (int, optional): day shift for regression. Defaults to 30.
        """
        self.df = df
        self.features = features
        self.target = target
        self.fitted = False

        self.model = LinearRegressor(transform=np.log, invtransform=np.exp)
        self.last_date = df.index[-1].strftime('%Y-%m-%d')
        self.current_price = df[target].values[-1]
        self.shift = shift

    def fit(self, prange=np.linspace(0.8, 1.2, 1001)):
        if self.df is None:
            raise ValueError(
                'No data available.')

        X, y = self.df[self.features], self.df[self.target]
        self.model.fit(X, y, shift=-self.shift)

        self.df['Prediction'] = self.model.predict(X)
        CalculateProbability(self.model, prange)

    def display(self, return_figures=False, xlabel='Price'):
        if not self.fitted:
            self.fit()

        # prob distribution
        title = f"R2={self.model.R2:.2f} (as of {self.last_date})"
        fig1 = VisualizeProbability(
            self.model, self.current_price, xlabel, title, False)

        # time series of predictions
        self.df = self.df.join(self.df.Prediction.shift(
            self.shift).rename('PredictedPrice'), how='outer')
        fig2 = px.line(self.df.ffill(), y=['PredictedPrice', 'ClosingPrice'])
        fig2.update_layout(xaxis_title='Time',
                           yaxis_title='Price', hovermode='x')
        if not return_figures:
            fig1.show(), fig2.show()
        else:
            return fig1, fig2

    def backtest(self, window):
        X, y = self.df[self.features], self.df[self.target]
        df_results = self.model.rolling_fit(
            X, y, shift=-self.shift, window=window)
        df_results['Observed'] = y
        # TODO: add PnL, convergence rate, etc.
        return df_results



from fredapi import Fred


class FredData(object):
    def __init__(self, api_key):
        """Class containing named shortcuts to useful FRED series.

        Args:
            api_key (string): FRED API key
        """
        self.client = Fred(api_key)

    def _get_series(series, name): return (lambda self, start=None,
                                           end=None: self.client.get_series(series, start, end).rename(name))

    CPI = _get_series('CPIAUCSL', 'CPI')
    Treasury10Y = _get_series('DGS10', 'Treasury10Y')
    InterestRate = _get_series('AMERIBOR', 'InterestRate')



import numpy as np

def shift(x, n, fill_value=None):
    """Shift 1D array by n elements

    Args:
        x (array): 1D array to be shifted
        n (int): Number of places to be shifted. Negative shifts element to the left.
        fill (float or None): Value used to fill the empty space created. If None then use the closest value. Default is None.
    
    Returns:
        shifted (array): Shifted array
    """
    if n == 0:
        return x
    res = np.roll(x, n)
    if n > 0:
        res[:n] = res[n] if fill_value is None else fill_value
    elif n < 0:
        res[n:] = res[n-1] if fill_value is None else fill_value
    return res

def fillna(x, fill_value=0):
    x = np.asarray(x)
    x[np.isnan(x)] = fill_value
    return x


import os
import requests
from fnmatch import fnmatch
from urllib.parse import urlsplit, urlunsplit
from bs4 import BeautifulSoup


def _try_get(url, ignore=False):
    response = None
    try:
        response = requests.get(url)
        if response.status_code != 200:
            raise RuntimeError(f"Failed to retrieve {url}.")
    except Exception as ex:
        if not ignore:
            raise ex
    return response


def download_links_on_page(page_url, download_dir, overwrite=False, regex=None, verbose=False, fname_keep_url_levels=1):
    """Download all links on a page

    Args:
        page_url (str): URL to the target page
        download_dir (str): Directory used to store the downloaded content
        overwrite (bool): Whether to re-download exisiting files. Default is False.
        regex (str or None, optional): UNIX filename pattern to filter links, e.g. "*.csv". 
            The match is performed via fnmatch(). If None download everything. Defaults to None.
        verbose (bool): Whether to print out the list of links to be downloaded. Defaults to False.
        fname_keep_url_levels (int): Number of URL levels to keep in filenames in case of conflict. Default is 1 (only base name).

    Returns:
        downloaded_files (List[str]): Paths to downloaded files
    """
    # check download directory
    if not os.path.exists(download_dir):
        os.makedirs(download_dir)

    response = _try_get(page_url)

    soup = BeautifulSoup(response.content, 'html.parser')
    links = [a['href'] for a in soup.find_all('a', href=True)]
    if not regex is None:
        links = filter(lambda s: fnmatch(s, regex), links)

    downloaded_files = []
    for link in links:
        if link.startswith('/'):
            split_url = urlsplit(page_url)
            link = f"{split_url.scheme}://{split_url.netloc}{link}"

        if verbose:
            print(f'Downloading {link}')
        basename = '_'.join(link.split('?')[0].rstrip('/').split('/')[-fname_keep_url_levels:])
        fname = f"{download_dir.rstrip('/')}/{basename}"
        if fname in downloaded_files:
            continue
        if overwrite or (not os.path.exists(fname)):
            response = _try_get(link, ignore=True)
            with open(fname, 'wb') as f:
                f.write(response.content)
        downloaded_files.append(fname)

    return downloaded_files



import pandas as pd
pd.options.mode.chained_assignment = None
import numpy as np
import plotly.express as px
from plotly.subplots import make_subplots
from scipy.stats import gaussian_kde


def CalculateUncertainty(model, use_percentage=True):
    if (not hasattr(model, 'ypred')) or (not hasattr(model, 'observed')):
        raise ValueError(
            'Model must contain "observed" values and "ypred" values.')
    if model.prediction is None or model.observed is None:
        raise ValueError('Model is not fitted.')

    if use_percentage:
        model.uncertainty = gaussian_kde(model.observed/model.ypred)
    else:
        model.uncertainty = gaussian_kde(model.observed - model.ypred)

    return model.uncertainty


def CalculateProbability(model, prange,  current_pred=None, use_percentage=True):
    if (not hasattr(model, 'uncertainty')) or (model.uncertainty is None):
        CalculateUncertainty(model, use_percentage)
    if current_pred is None:
        current_pred = model.prediction[-1]

    values = (prange*current_pred) if use_percentage else (prange+current_pred)

    model.prob = pd.DataFrame(model.uncertainty(
        prange), index=values, columns=['probability'])
    model.prob['pvalue'] = [
        model.uncertainty.integrate_box_1d(p, np.inf) for p in prange]
    model.prob = model.prob.round(4)

    return model.prob


def VisualizeProbability(model, current_value=None, xlabel='Value', title='', show_figure=False):
    if (not hasattr(model, 'prob')) or (model.prob is None):
        raise ValueError(
            'Probabilities has not been calculated for model. Call CalculateProbability(model) first.')

    fig = make_subplots(specs=[[{"secondary_y": True}]])
    fig1 = px.line(model.prob, y='pvalue')
    fig2 = px.line(model.prob, y='probability')
    fig2.update_traces(yaxis="y2")
    fig.add_traces(fig1.data + fig2.data)
    fig.add_vline(current_value, line_color='red',
                   annotation_text='Current Value')
    fig.layout.xaxis.title = xlabel
    fig.layout.yaxis.title = "p-value"
    fig.layout.yaxis2.title = "Probability"
    fig.update_layout(title=title, hovermode='x')
    fig.for_each_trace(lambda t: t.update(line=dict(color=t.marker.color)))

    if show_figure:
        fig.show()
    return fig
        
def ConvergenceTrades(data, observed_column='Observed', pred_column='Prediction'):
# FIXME: add percentage_based

    df = data[[observed_column,pred_column,'R2']].dropna()
    df.loc[:,'Difference'] = df.eval(f'{pred_column}-{observed_column}')

    df_underprice = df.query('Difference >= 0')
    df_underprice.loc[:,'Time'] = df_underprice.index
    df_overprice = df.query('Difference < 0')
    df_overprice.loc[:,'Time'] = df_overprice.index

    df_short = pd.merge_asof(df_overprice, df_underprice, left_index=True, right_index=True, suffixes=('1','2'), direction='forward')
    df_long = pd.merge_asof(df_underprice, df_overprice, left_index=True, right_index=True, suffixes=('1','2'), direction='forward')

    df_short.loc[:,'Side'] = -1
    df_long.loc[:,'Side'] = +1

    df_trades = pd.concat([df_short,df_long]).sort_index().dropna()
    df_trades.loc[:,'PnL'] = df_trades.eval(f'Side*({observed_column}2-{observed_column}1)')
    df_trades.loc[:,'Days'] = (df_trades['Time2']-df_trades['Time1']).dt.days
    df_trades.loc[:,'Speed'] = df_trades.eval('PnL/Days')
    df_trades.loc[:,'AbsDifference'] = df_trades.eval('abs(Difference1)')

    return df_trades


"""Collection of smoothing methods not found in standard libraries
"""

import numpy as np

def anisotropic_smoothing1d(y0, dx, steps, K, dt=1e-6, diffusion_type='cauchy'):
    """1D anisotropic diffusion smoothing
    References:
        1. https://www.cs.sfu.ca/~stella/papers/blairthesis/main/node25.html
        2. https://en.wikipedia.org/wiki/Anisotropic_diffusion

    Args:
        y0 (array): Function to be smoothed.
        dx (float or array): Grid size.
        steps (int): Number of integration steps. More steps results more smoothing.
        K (float): Width of the diffusion kernel
        dt (float, optional): Time step. Must be small for convergence. Defaults to 1e-6.
        diffusion_type (str, optional):Type of diffusion kernel. Defaults to 'cauchy'.

    Returns:
        yt (array): Smoothed function.
    """
    fcauchy = lambda x: 1/(1+(x/K)**2)
    fgauss = lambda x: np.exp(-(x/K)**2)
    f = fgauss if diffusion_type == 'gauss' else fcauchy

    yt = y0.copy()
    for _ in range(steps):
        fdiff = np.diff(yt,append=yt[0]) # forward difference
        bdiff = np.diff(yt[::-1],append=yt[-1])[::-1] # backward difference
        flow_right = fdiff*f(abs(fdiff)/dx) / dx**2
        flow_left =  bdiff*f(abs(bdiff)/dx) / dx**2
        yt += dt*(flow_right+flow_left + (y0-yt))
    return yt


import numpy as np


def total_variation(y, x):
    return abs(np.diff(y)/np.diff(x)).sum() * np.diff(x).mean()


def integral_rmse(dydx, y, x):
    """Calculate the RMSE between y estimated by integrating `dydx` and `y`

    Args:
        dydx (array): Estimated derivative of y
        y (array): Observed values of y
        x (array): Independent variable

    Raises:
        ValueError: x must be sorted

    Returns:
        RMSE (float): RMSE error between integral of `dydx` and `y`
    """
    if not np.all(np.diff(x) >= 0):
        raise ValueError('x is not sorted.')
    y_est = [y[0]]
    for i in range(1, len(x)):
        y_est.append(y[0]+np.trapz(dydx[:i+1], x[:i+1]))
    return np.sqrt(np.mean(np.subtract(y, y_est)**2))



# %%
from thetadata import ThetaClient
import pandas as pd
import requests


class OptionsData(object):
    def __init__(self, username, password):
        """Wrapper for ThetaData client to obtain live options quotes

        Args:
            username (str): ThetaData username
            password (str): ThetaData password
        """
        self.client = ThetaClient(username=username, passwd=password)

    def get_expirations(self, symbol):
        with self.client.connect():
            return self.client.get_expirations(symbol)

    def get_front_silo(self, symbol, dte=False):
        today = pd.to_datetime(pd.Timestamp.today().date())
        silos = self.get_expirations(symbol)
        expiration = min(silos[silos >= today])
        if dte:
            # TODO: calculate current hour til 4pm EST, subtract from dte
            return (expiration - today).days
        return expiration

    def get_snapshot(self, symbol, expiration=None):
        with self.client.connect():
            if expiration is None:  # Use front silo if not specified
                expiration = self.get_front_silo(symbol)
            api_url = f"http://127.0.0.1:25510/bulk_snapshot/option/quote?root={symbol}&exp={expiration.strftime('%Y%m%d')}"
            entries = []
            response = requests.get(api_url)
            for tick in response.json()['response']:
                strike = tick['contract']['strike']
                itype = tick['contract']['right']
                _, bidsize, _, bid, _, asksize, _, ask, _, _ = tick['tick']
                mprice = (bidsize*ask+asksize*bid) / \
                    (bidsize+asksize) if (bidsize+asksize) > 0 else 0
                entries.append({'Strike': strike/1000, 'Spread': round(ask -
                               bid, 2), 'MPrice': round(mprice, 2), 'IType': itype})
        return pd.DataFrame(entries).drop_duplicates()



import pandas as pd
import numpy as np
from derivative import dxdt


def _implied_dist(strikes, prices):
    d1 = dxdt(prices, strikes, kind="kalman", alpha=1000)
    d2 = dxdt(d1, strikes, kind="kalman", alpha=1000)

    # zero out the tails
    dist = pd.Series(d2, index=strikes)
    atm = dist.idxmax()
    otm = dist[atm:]
    itm = dist[:atm]
    max_strike, min_strike = otm.where(
        otm <= 0).first_valid_index(), itm.where(itm <= 0).last_valid_index()
    res = dist[min_strike:max_strike].iloc[1:-1]

    return res / np.trapz(res, x=res.index)


def _expectation(dist):
    return np.trapz(dist*dist.index, x=dist.index)


class OptionsImpliedPrice(object):
    def __init__(self, options_data, interest_rate):
        self.data = options_data
        self.interest_rate = interest_rate
        self.ert = {}

    def __getitem__(self, symbol):
        df = self.data.get_snapshot(symbol)
        calls = df.query('IType=="C"').sort_values('Strike')
        puts = df.query('IType=="P"').sort_values('Strike')
        if symbol not in self.ert:
            self.ert[symbol] = np.exp(-self.interest_rate *
                                      self.data.get_front_silo(symbol, dte=True)/256)
        call_implied = _expectation(_implied_dist(
            calls.Strike.values, calls.MPrice.values))*self.ert[symbol]
        put_implied = _expectation(_implied_dist(
            puts.Strike.values, puts.MPrice.values))*self.ert[symbol]
        return sorted([call_implied.round(2), put_implied.round(2)])



from maldives.regression import LinearRegressor
from maldives.api.usda import CattleWeeklyProduction
from maldives.api import FredData
from maldives.backtest.utils import CalculateProbability, VisualizeProbability

import yfinance as yf
import pandas as pd
import numpy as np
import plotly.express as px


class CattlePriceModel(object):
    def __init__(self, shift=4):
        self.df = None
        self.fitted = False

        self.model = LinearRegressor(transform=np.log, invtransform=np.exp)
        self.last_date = None
        self.current_price = np.nan
        self.prob = None
        self.uncertainty = None
        self.shift = shift

    def load_data(self, fred_api_key, start=pd.Timestamp.today()-pd.tseries.offsets.BDay(300), end=pd.Timestamp.today()):
        if type(start) is int:
            start = end - pd.tseries.offsets.BDay(start)

        # load gold futures prices
        days_back_required = (pd.Timestamp.today()-start).round('D').days
        df_cattle = CattleWeeklyProduction().load_data(days_back_required/5+4)

        cattle = yf.Ticker("LE=F").history(start=start, end=end)
        cattle['Date'] = pd.to_datetime(cattle['Date'])
        cattle['ClosingPrice'] = cattle['Close']
        cattle = cattle.set_index('Date')[['ClosingPrice']]
        self.current_price = cattle.ClosingPrice.values[-1]

        # load cpi and treasury yield
        fred = FredData(fred_api_key)
        cpi = fred.CPI()

        df = pd.merge_asof(cattle, cpi, left_index=True,
                           right_index=True, direction='nearest')
        df = pd.merge_asof(df_cattle, df, left_index=True,
                           right_index=True, direction='forward')

        df = df.dropna().loc[start.date():end.date()]
        self.df = df
        self.last_date = self.df.index[-1].strftime('%Y-%m-%d')

    def fit(self):
        if self.df is None:
            raise ValueError(
                'No data available. Please call load_data() before calling fit().')

        X, y = self.df[['Steers', 'Heifers', 'CPI', 'Cattle']], self.df['ClosingPrice']
        self.model.fit(X, y, shift=-self.shift)

        self.df['Prediction'] = self.model.predict(X)
        CalculateProbability(self.model, np.linspace(0.8, 1.2, 1001))

    def display(self, return_figures=False):
        if not self.fitted:
            self.fit()

        # prob distribution
        title = f"R2={self.model.R2:.2f} (as of {self.last_date})"
        fig1 = VisualizeProbability(
            self.model, self.current_price, 'Live cattle Price', title, False)

        # time series of predictions
        self.df = self.df.join(self.df.Prediction.shift(
            self.shift).rename('PredictedPrice'), how='outer')
        fig2 = px.line(self.df.ffill(), y=['PredictedPrice', 'ClosingPrice'])
        fig2.update_layout(xaxis_title='Time',
                           yaxis_title='Price', hovermode='x')
        if not return_figures:
            fig1.show(), fig2.show()
        else:
            return fig1, fig2

    def backtest(self, window):
        data = self.df[['Steers', 'Heifers', 'CPI', 'Cattle', 'ClosingPrice']].dropna().drop_duplicates()
        X, y = data[['Steers', 'Heifers', 'CPI', 'Cattle']], data.dropna()[
            'ClosingPrice']
        df_results = self.model.rolling_fit(
            X, y, shift=-self.shift, window=window)
        df_results['Observed'] = data.ClosingPrice
        # TODO: add PnL, convergence rate, etc.
        return df_results



from ray import tune


class TunedFunction(object):
    def __init__(self, function, obj_function, param_space, **kwargs):
        """Create a wrapper around a function call that takes some hyperparameters as arguments.
        The tuner is a ray tuner.

        Args:
            function (Function): Main function to be called. 
            obj_function (Function): Objective function to evaluate the output from main function. 
                The objective function should have the same signature as the main function, plus an additional "output" args accepting the evaluated output of the main function. 
            param_space (dict[ray.tune]): Dictionary specifying the search space (use ray.tune). This is passed into a ray tuner. See ray.io for docs.

        Any additional arguments will be passed into the ray tuner. At minimum you can pass in tune_config=tune.TuneConfig(num_samples=10, metric='score', mode='min').

        Example::
            from maldives.env import *

            xx = np.linspace(0,1,101)
            def sinefunction(x,a=1):
                return np.sin(a*x)

            search_space = {'a':tune.uniform(0,10)}

            # we want to target a=3
            def loss(x,output=None, **kwargs):
                return {'loss':sum((output-np.sin(3*x))**2)}

            tune_config=tune.TuneConfig(num_samples=100, metric='loss', mode='min')
            tunedsine = TunedFunction(sinefunction, loss, search_space, run_config=air.RunConfig(verbose=0), tune_config=tune_config)

            plt.plot(tunedsine(xx))
            plt.plot(sinefunction(xx,3))
            plt.plot(tunedsine(xx, a=10)) # we can still explicitly override tuned parameter
        """
        self._function = function
        self._objective = obj_function
        self._params = param_space
        self._tuner_args = kwargs
        self.best_config = None
        self.tuner = None

    def __call__(self, *args, **kwargs):
        """Call the main function. We use tuned values for arguments that are in the search space and not specified.

        Returns:
            _type_: _description_
        """
        if self.best_config is None:
            self.optimize(*args, **kwargs)

        # fill in values from best config which are not explicitly specified.
        for k, v in self.best_config.items():
            if k not in kwargs:
                kwargs[k] = v

        return self._function(*args, **kwargs)

    def optimize(self, *args, **kwargs):
        """Optimize config based on passed inputs.
            This has the same signature as the main function. Arguments that are in the search space and not specified are tuned.
        """
        # if a config is explicitly specified, we do not want to optimize it.
        search_space = {k: v for k, v in self._params.items()
                        if k not in kwargs}

        def objective(config):
            output = self._function(*args, **kwargs, **config)
            return self._objective(*args, output=output, **kwargs, **config)

        self.tuner = tune.Tuner(
            objective, param_space=search_space, **self._tuner_args)
        results = self.tuner.fit()
        self.best_config = results.get_best_result().config



"""Collection of numerical methods not found in standard libraries
"""

import numpy as np
from scipy.special import comb
from functools import cache

import maldives.utils as md


def robust_deriv(y, x, order=1, N=5, fill_value=None):
    """Low noise finite differentiator
    Suppresses high-frequency contribution to the numerical derivative.
    Requires tuning size of stencil in order to balance accuracy and smoothness.

    References:
        1. http://www.holoborodko.com/pavel/numerical-methods/numerical-derivative/smooth-low-noise-differentiators/

    Args:
        y (array): Function to take derivative on.
        x (array): Independent variable.
        order (int, optional): Order of derivative. Default is first-derivative.
        N (int, optional): Size of the stencil. Higher is more accurate but requires more data. Must be odd number. Default is 5.
        fill_value (float or None, optional): Value used to fill the derivative at boundaries. If None then calculate derivates by padding y with zeros.
    """
    if order not in [1, 2]:
        raise NotImplementedError(
            'Order must be 1 or 2. Higer-orders are not implemented.')
    if N % 2 != 1:
        raise ValueError('N must be an odd integer.')

    m = (N-3) // 2
    M = (N-1) // 2

    @cache
    def s(k):
        if k > M:
            return 0
        elif k == M:
            return 1
        return ((2*N-10)*s(k+1)-(N+2*k+3)*s(k+2))/(N-2*k-1)

    res = []

    NaNs = [np.nan] * (2*M)
    x_padded = np.concatenate((x.copy(), NaNs))
    y_padded = np.concatenate((y.copy(), NaNs))
    if fill_value is None:
        x_padded = md.fillna(x_padded)
        y_padded = md.fillna(y_padded)

    for i in range(len(y)):
        fp = 0
        for k in range(1, M+1):
            dx = (x_padded[i+k]-x_padded[i-k])/(2*k)
            if dx == 0:
                continue
            if order == 1:
                ck = (comb(2*m, m-k+1)-comb(2*m, m-k-1)) / 2**(N-2)
                fp += ck*(y_padded[i+k]-y_padded[i-k]) / dx
            elif order == 2:
                fp += (y_padded[i+k]-2*y_padded[i] +
                       y_padded[i-k])/dx**2*s(k) / 2**(N-3)
        res.append(fp)
    return (res if fill_value is None else md.fillna(res, fill_value))



import matplotlib.pyplot as plt


def plot(*args, **kwargs):
    """Wrapper for pyplot with better defaults.
    """
    fig, ax = plt.subplots(dpi=150)
    ax.plot(*args, **kwargs)
    ax.grid()
    if 'label' in kwargs:
        ax.legend()
    return fig, ax


def multiplot(xx, ydict, styling={}, **kwargs):
    """Plot multiple y values on the same axis.

    Args:
        xx (array): x-axis values.
        ydict (dict[array]): Dictionary of y-values with the keys being the legend labels.
        styling (dict[dict]): Dictionary of additional styling arguments for each y-values.
    """
    fig, ax = None, None
    for k, v in ydict.items():
        if ax is None:
            fig, ax = plot(xx, v, label=k, **styling.get(k, {}), **kwargs)
            continue
        ax.plot(xx, v, **styling.get(k, {}), label=k)
    ax.legend()
    return fig, ax
